;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;;
;;; 01_stacks.lisp - Zásobníkové programování
;;;
;;; Jazyk pro myšlený zásobníkový stroj.
;;;

;; Je třeba načíst soubor 01_stacks_1.lisp
 
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;;
;;; fáze 2 - vazby
;;;

;;
;; Slovník vazeb (proměnná je definována v předchozím souboru)
;;

(defvar *bnd*)
(setf *bnd* `((pi . ,pi)))

;; Vytvoření vazby symbolu na hodnotu a její zrušení
(defprim :bind
  (dict-push (pop *rslt*) (pop *rslt*) *bnd*))

(defprim :unbind
  (pop *bnd*))

;; Zjištění hodnoty symbolu
(defprim :val
  (push (dict-find (pop *rslt*) *bnd*)
        *rslt*))

#|

(execute 3 'x :bind 4 'y :bind 'x :val 'y :val :- :unbind :unbind)

;; obsah kruhu daného poloměru:
(execute :def :carea :dup 'pi :val :* :*)
(execute 10 :carea)

|#

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;;
;;; fáze 3 - podprogramy
;;;

#|

Podprogram je seznam s kódem zásobníkového stroje. 

|#

;; Volání podprogramu (musí být na vrcholu datového zásobníku)
(defprim :clsub
  (push *exec* *ret*)
  (setf *exec* (pop *rslt*)))

;; Návrat z podprogramu
(defprim :ret
  (setf *exec* (pop *ret*)))



;; Definice podprogramů na aritmetické operace.
;; Můžeme používat místo slov (hlavně na další přednášce)
(execute '(:+ :ret) '+ :bind)
(execute '(:- :ret) '- :bind)
(execute '(:* :ret) '* :bind)
(execute '(:/ :ret) '/ :bind)
(execute '(:= :ret) '= :bind)

#|

;; Příklady 

(execute 2 3 '+ :val :clsub)
(execute 2 3 '+ :val :clsub 4 '* :val :clsub)

;; faktoriál jako podprogram: 

(execute '(:dup 0 := :if :dup 1 :- fact :val :clsub :* :else :drop 1 :then :ret)
         'fact
         :bind)

(execute 4 'fact :val :clsub)

|#

#|
Podprogram může obsahovat libovolný kód, může si tedy dělat se zásobníky, co chce.
Můžeme jej ale přiblížit procedurám (funkcím) z vyšších programovacích jazyků
využitím parametrů. Například kód procedury s jedním parametrem by mohl vypadat takto:

x :bind ... kód využívající parametr x ... :unbind

Výsledek je o něco čitelnější. Například faktoriál:

(execute '(n :bind
               n :val 0 := :if n :val n :val 1 :- fact :val :clsub :* :else
                               1 :then
             :unbind
           :ret)
           'fact
           :bind)
 
(execute 5 'fact :val :clsub)

;; Více parametrů je třeba navazovat v obráceném pořadí.
;; např. výpočet diskriminantu kvadratické rovnice s koeficienty
;; a, b, c:

(execute '(c :bind 
           b :bind
           a :bind
              b :val b :val :* 4 a :val c :val :* :* :-
           :unbind
           :unbind
           :unbind
           :ret)
         'sqdiscr :bind)

(execute 6 11 -2 'sqdiscr :val :clsub)

|#



