;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;;
;;; 02_stacks_5.lisp - Zásobníkové programování
;;;
;;; Jazyk pro myšlený zásobníkový stroj.
;;;

;;
;; REPL -- Scheme
;;

(defun scheme-repl ()
  (loop (format t "~%scheme > ") 
        (handler-case (print (evaluate (read)))
          (error (err) (format t "Chyba: ~a" err)))))

;;
;; Další procedury pro porovnávání
;;

(execute '(:> :ret)
         '> :bind)

(execute '(:< :ret)
         '< :bind)

(execute '(:>= :ret)
         '>= :bind)

(execute '(:<= :ret)
         '<= :bind)

;;
;; Další procedury pro práci se seznamy
;;

(evaluate '(define rest cdr))

(evaluate '(define first car))

(evaluate '(define second (lambda (lst)
                           (first (cdr lst)))))

(evaluate '(define third (lambda (lst)
                          (second (cdr lst)))))

(evaluate '(define fourth (lambda (lst)
                          (third (cdr lst)))))

(evaluate '(define fifth (lambda (lst)
                           (fourth (cdr lst)))))


(evaluate '(define mapcar (lambda (fun lst)
                            (if lst
                                (cons (fun (car lst))
                                      (mapcar fun (cdr lst)))
                              nil))))


#|
(evaluate '(mapcar (lambda (x) (+ x 1)) '(1 2 3)))
|#

;;
;; speciální operátor set! na nastavení hodnoty proměnné
;;


(execute 'set! 'specials :val :cons 'specials :set!)
(execute '(:split :split :drop eval :val :clsub :swap :set! :ret)
         'set! :bind)
                           
#|
((lambda (x) 
   (print x) 
   (set! x 2) 
   x)
 4)
|#


;;
;; Přidáme makra
;;


(execute nil 'macros :bind)

;; Vyhodnocení slož. výrazu --- eval-comp. 
;; ((op a1 ... an) -- hodnota)
(execute '(:split :swap              ;(a1 ... an) op
           :dup specials :val        ;(a1 ... an) op op (quote ...)
                find :val :clsub     ;(a1 ... an) op nil/t
                :if                  ;op není speciální
                    :dup macros :val ;(a1 ... an) op op (...)
                    find :val :clsub ;(a1 ... an) op nil/t
                    :if              ;op není makro 
                                     ;spustíme pomocnou proceduru eval-app
                                     ;na vyhodnocení obecné aplikace
                        eval-app :val :clsub :else
                                     ;spustíme proceduru uloženou v op a znovu vyhodnotíme
                        :val :clsub eval :val :clsub :then :else
                                     ;spustíme proceduru uloženou v op 
                    :val :clsub :then  
           :ret)
         'eval-comp :set!)





;; Makro list
(execute 'list 'macros :val :cons 'macros :set!)

(evaluate `(define list (lambda (args)
                          (if args
                              (cons 'cons (cons (car args) (cons (cons 'list (cdr args)) nil)))
                            '()))))


;; Makro defmac
(execute 'defmac 'macros :val :cons 'macros :set!)

(evaluate '(define defmac (lambda (args)
                            (list 'begin
                                  (list 'set! 'macros (list 'cons (list 'quote (first args)) 'macros))
                                  (list 'define (first args) (second args))))))

;; Makro let
(evaluate '(defmac let (lambda (args)
                         (cons
                          (cons 'lambda
                                (cons
                                 (mapcar first (first args))
                                 (rest args)))
                          (mapcar second (first args))))))

#|
(evaluate '(let ((a (+ 1 1))
                 (b (* 2 2)))
             (+ a b)))
|#



;; Speciální operátor do-while
(execute 'do-while 'specials :val :cons 'specials :set!)

(execute '(begin :swap :cons :lbl :dup eval :val :clsub  :cjmp :drop nil :ret) 'do-while :bind)

#|
;; Scheme

(let ((a 5))
  (do-while
     (print a)
     (set! a (- a 1))
     (> a 0)))
|#

;; Makro while
(evaluate '(defmac while (lambda (args)
                           (list 'if (first args)
                                 (cons 'do-while (append (rest args) (list (first args))))
                                 nil))))

#|
;; Scheme

(let ((a 5))
  (while (> a 0)
         (print a)
         (set! a (- a 1))))
|#
;; Generování symbolů pro makra


(defprim :gensym
  (push (gensym (pop *rslt*)) *rslt*))

#|
(execute "A" :gensym)
|#

(execute '(:gensym :ret) 'gensym :bind)

#|
(evaluate '(gensym "A"))
|#

;; Makro dotimes
(evaluate '(defmac dotimes 
                   (lambda (args)
                     (let ((var (first (first args)))
                           (count-form (second (first args)))
                           (body (rest args))
                           (count (gensym "COUNT")))
                       (list 'let (list (list var 0)
                                        (list count count-form))
                             (cons 'while 
                                   (cons (list '< var count)
                                         (append body (list (list 'set! var (list '+ var 1)))))))))))

#|
;; Scheme

(dotimes (i 10) 
  (print i))
|#





