;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;;
;;; 06_baton.lisp - Štafetový kolík
;;;


;; Makro cond
(def-scheme-macro cond (&rest args)
  (if args
      (let ((condition (caar args))
            (body (cdar args))
            (rest (cdr args)))
        `(if ,condition
             (begin ,@body)
           (cond ,@rest)))
    nil))

#|
(let ((a 1))
  (cond
   ((= a 0)
    (print 1)
    (print 2))
   ((= a 1)
    (print 3)
    (print 4))
   (t
    (print 5)
    (print 6))))
|#

;; Inkrementace a dekrementace prvku vektoru:
(compiled-evaluate '(define vinc! (lambda (vector i)
                                    (vset! vector i (+ (vref vector i) 1)))))

(compiled-evaluate '(define vdec! (lambda (vector i)
                                    (vset! vector i (- (vref vector i) 1)))))

#|
(let ((v1 #(1 1)))
  (vinc! v1 0)
  (vdec! v1 1)
  v1)
|#

;; Inkrementace a dekrementace prvku seznamu:
(compiled-evaluate '(define linc! (lambda (lst i)
                                    (lset! lst i (+ (lref lst i) 1)))))

(compiled-evaluate '(define ldec! (lambda (lst i)
                                    (lset! lst  i (- (lref lst i) 1)))))

#|
(let ((l (list 1 2)))
  (linc! l 0)
  (ldec! l 1)
  l)
|#

;; Kolík je semafor s výchozí hodnotou jedna.
(compiled-evaluate '(define make-baton 
                            (lambda ()
                              (sem 1))))

;; Skupina je seznam (baton count semaphore),
;; kde baton je kolík, semaphore je semafor a count je počet procesů čekajících na semaforu.
(compiled-evaluate '(define make-baton-group 
                            (lambda (baton)
                              (list baton 0 (sem 0)))))

;; Selektory:
(compiled-evaluate '(define get-bg-baton first))
(compiled-evaluate '(define get-bg-count second))
(compiled-evaluate '(define get-bg-semaphore third))

;; Inkrementace a dekrementace počtu procesů:
(compiled-evaluate '(define bg-inc! (lambda (bg)
                                      (linc! bg 1))))

(compiled-evaluate '(define bg-dec! (lambda (bg)
                                      (ldec! bg 1))))
                            
;; Operace skupiny:
(compiled-evaluate '(define baton-waiting-p 
                            (lambda (bg)
                              (< 0 (get-bg-count bg)))))
                  
(compiled-evaluate '(define baton-signal 
                            (lambda (bg)
                              (bg-dec! bg)
                              (v (get-bg-semaphore bg)))))

(compiled-evaluate '(define baton-wait 
                            (lambda (bg)
                              (bg-inc! bg)
                              (v (get-bg-baton bg))
                              (p (get-bg-semaphore bg)))))


